import re

from clipped.utils.dates import parse_datetime

# pylint:disable=anomalous-backslash-in-string

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S %Z"  # noqa
ISO_DATETIME_REGEX = re.compile(  # noqa
    r"([0-9]+)-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01])[Tt]"
    r"([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.[0-9]+)?"
    r"(([Zz])|([\+|\-]([01][0-9]|2[0-3]):[0-5][0-9]))\s?"
)
DATETIME_REGEX = re.compile(  # noqa
    r"\d{2}(?:\d{2})?-\d{1,2}-\d{1,2}\s\d{1,2}:\d{1,2}:\d{1,2}\s\w+\s?"
)


def timestamp_search_regex(regex, log_line):
    log_search = regex.search(log_line)
    if not log_search:
        return log_line, None

    ts = log_search.group()
    ts = parse_datetime(ts)

    return re.sub(regex, "", log_line), ts



import os
import uuid

from argparse import Namespace
from typing import Any, Dict, List, Optional, Union

import packaging

from polyaxon._env_vars.keys import ENV_KEYS_RUN_INSTANCE
from polyaxon.client import RunClient
from traceml import tracking
from traceml.exceptions import TracemlException

try:
    import pytorch_lightning as pl

    NEW_PL_VERSION = packaging.version.parse(pl.__version__)

    if NEW_PL_VERSION < packaging.version.parse("1.7"):
        from pytorch_lightning.loggers.base import LightningLoggerBase as Logger
        from pytorch_lightning.loggers.base import rank_zero_experiment
    else:
        from pytorch_lightning.loggers.logger import Logger, rank_zero_experiment

    if NEW_PL_VERSION < packaging.version.parse("1.9"):
        from pytorch_lightning.utilities.logger import (
            _add_prefix,
            _convert_params,
            _flatten_dict,
            _sanitize_callable_params,
        )
    else:
        from lightning_fabric.utilities.logger import (
            _add_prefix,
            _convert_params,
            _flatten_dict,
            _sanitize_callable_params,
        )
    from pytorch_lightning.utilities.model_summary import ModelSummary
    from pytorch_lightning.utilities.rank_zero import rank_zero_only, rank_zero_warn
except ImportError:
    raise TracemlException("PytorchLightning is required to use the tracking Callback")


class Callback(Logger):
    LOGGER_JOIN_CHAR = "_"

    def __init__(
        self,
        owner: Optional[str] = None,
        project: Optional[str] = None,
        run_uuid: Optional[str] = None,
        client: RunClient = None,
        track_code: bool = True,
        track_env: bool = True,
        refresh_data: bool = False,
        artifacts_path: Optional[str] = None,
        collect_artifacts: Optional[str] = None,
        collect_resources: Optional[str] = None,
        is_offline: Optional[bool] = None,
        is_new: Optional[bool] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        end_on_finalize: bool = False,
        prefix: str = "",
    ):
        super().__init__()
        self._owner = owner
        self._project = project
        self._run_uuid = run_uuid
        self._client = client
        self._track_code = track_code
        self._track_env = track_env
        self._refresh_data = refresh_data
        self._artifacts_path = artifacts_path
        self._collect_artifacts = collect_artifacts
        self._collect_resources = collect_resources
        self._is_offline = is_offline
        self._is_new = is_new
        self._name = name
        self._description = description
        self._tags = tags
        self._end_on_finalize = end_on_finalize
        self._prefix = prefix
        self._experiment = None

    @property
    @rank_zero_experiment
    def experiment(self) -> tracking.Run:
        if self._experiment:
            return self._experiment
        tracking.init(
            owner=self._owner,
            project=self._project,
            run_uuid=self._run_uuid,
            client=self._client,
            track_code=self._track_code,
            track_env=self._track_env,
            refresh_data=self._refresh_data,
            artifacts_path=self._artifacts_path,
            collect_artifacts=self._collect_artifacts,
            collect_resources=self._collect_resources,
            is_offline=self._is_offline,
            is_new=self._is_new,
            name=self._name,
            description=self._description,
            tags=self._tags,
        )
        self._experiment = tracking.TRACKING_RUN
        return self._experiment

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]):
        params = _convert_params(params)
        params = _flatten_dict(params)
        params = _sanitize_callable_params(params)
        self.experiment.log_inputs(**params)

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None):
        assert rank_zero_only.rank == 0, "experiment tried to log from global_rank != 0"
        metrics = _add_prefix(metrics, self._prefix, self.LOGGER_JOIN_CHAR)
        self.experiment.log_metrics(**metrics, step=step)

    @rank_zero_only
    def log_model_summary(self, model: "pl.LightningModule", max_depth: int = -1):
        summary = str(ModelSummary(model=model, max_depth=max_depth))
        rel_path = self.experiment.get_outputs_path("model_summary.txt")
        with open(rel_path, "w") as f:
            f.write(summary)
        self.experiment.log_file_ref(
            path=rel_path, name="model_summary", is_input=False
        )

    @property
    def save_dir(self) -> Optional[str]:
        return self.experiment.get_outputs_path()

    @rank_zero_only
    def finalize(self, status: str):
        if self._end_on_finalize:
            self.experiment.end()
            self._experiment = None

    def _set_run_instance_from_env_vars(self, force: bool = False):
        """Tries to extract run info from canonical env vars"""
        run_instance = os.getenv(ENV_KEYS_RUN_INSTANCE)
        if not run_instance:
            return

        parts = run_instance.split(".")
        if len(parts) != 4:
            return

        if not self._name or force:
            self._name = parts[2]
        if not self._run_uuid or force:
            self._run_uuid = parts[-1]

    @property
    def name(self) -> str:
        if self._experiment is not None and self._experiment.run_data.name is not None:
            return self.experiment.run_data.name

        if not self._name:
            self._set_run_instance_from_env_vars()

        if self._name:
            return self._name

        return "default"

    @property
    def version(self) -> str:
        if self._experiment is not None and self._experiment.run_data.uuid is not None:
            return self.experiment.run_data.uuid

        if not self._run_uuid:
            self._set_run_instance_from_env_vars()

        if self._run_uuid:
            return self._run_uuid

        return uuid.uuid4().hex



from typing import TYPE_CHECKING, Any

from traceml import tracking
from traceml.exceptions import TracemlException
from traceml.integrations.tensorboard import Logger

try:
    import tensorflow as tf
except ImportError:
    raise TracemlException("tensorflow is required to use the tracking Callback")

try:
    from tensorflow.train import SessionRunHook  # noqa
except ImportError:
    raise TracemlException("tensorflow is required to use the tracking Callback")


if TYPE_CHECKING:
    from traceml.tracking import Run


class Callback(SessionRunHook):
    def __init__(
        self,
        summary_op: Any = None,
        steps_per_log: int = 1000,
        run: "Run" = None,
        log_image: bool = False,
        log_histo: bool = False,
        log_tensor: bool = False,
    ):
        self._summary_op = summary_op
        self._steps_per_log = steps_per_log
        self.run = tracking.get_or_create_run(run)
        self._log_image = log_image
        self._log_histo = log_histo
        self._log_tensor = log_tensor

    def begin(self):
        if self._summary_op is None:
            self._summary_op = tf.summary.merge_all()
        self._step = -1

    def before_run(self, run_context):
        self._step += 1
        return tf.train.SessionRunArgs({"summary": self._summary_op})

    def after_run(self, run_context, run_values):
        if self._step % self._steps_per_log == 0:
            Logger.process_summary(
                run_values.results["summary"],
                run=self.run,
                log_image=self._log_image,
                log_histo=self._log_histo,
                log_tensor=self._log_tensor,
            )



from traceml.events.paths import (
    get_asset_path,
    get_event_assets_path,
    get_event_path,
    get_resource_path,
)
from traceml.events.schemas import (
    LoggedEventListSpec,
    LoggedEventSpec,
    V1Event,
    V1EventArtifact,
    V1EventAudio,
    V1EventChart,
    V1EventChartKind,
    V1EventConfusionMatrix,
    V1EventCurve,
    V1EventCurveKind,
    V1EventDataframe,
    V1EventHistogram,
    V1EventImage,
    V1EventModel,
    V1Events,
    V1EventSpan,
    V1EventVideo,
)



from typing import Optional

from clipped.utils.paths import copy_file_path

from traceml.events import V1EventDataframe

try:
    import numpy as np
except ImportError:
    np = None


def dataframe_path(
    from_path: str,
    asset_path: str,
    content_type: Optional[str] = None,
    asset_rel_path: Optional[str] = None,
) -> V1EventDataframe:
    copy_file_path(from_path, asset_path)
    return V1EventDataframe(
        path=asset_rel_path or asset_path, content_type=content_type
    )



from traceml import tracking


def callback(run=None):
    run = tracking.get_or_create_run(run)

    def _callback(env):
        res = {}
        for data_name, eval_name, value, _ in env.evaluation_result_list:
            key = data_name + "-" + eval_name
            res[key] = value
        run.log_metrics(step=env.iteration, **res)

    return _callback



import datetime
import json

from collections import namedtuple
from typing import Any, Dict, List, Mapping, Optional, Union

from clipped.compact.pydantic import StrictStr, root_validator
from clipped.config.parser import ConfigParser
from clipped.config.schema import skip_partial
from clipped.types.uuids import UUIDStr
from clipped.utils.dates import parse_datetime
from clipped.utils.enums import PEnum
from clipped.utils.np import sanitize_np_types
from clipped.utils.strings import validate_file_or_buffer
from clipped.utils.tz import now

from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._schemas.lifecycle import V1StatusCondition, V1Statuses
from traceml.artifacts.enums import V1ArtifactKind
from traceml.logger import logger


class SearchView(str, PEnum):
    ANY = "any"
    RUNS = "runs"
    SELECTION = "selection"
    ANALYTICS = "analytics"
    COMPONENTS = "components"
    MODELS = "models"
    ARTIFACTS = "artifacts"
    PROJECTS = "projects"


class V1EventSpanKind(str, PEnum):
    LLM = "llm"
    CHAIN = "chain"
    AGENT = "agent"
    TOOL = "tool"
    EMBEDDING = "embedding"
    RETRIEVER = "retriever"


class V1EventSpan(BaseSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    started_at: Optional[datetime.datetime]
    finished_at: Optional[datetime.datetime]
    status: Optional[V1Statuses]
    status_conditions: Optional[List[V1StatusCondition]]
    metadata: Optional[Dict[str, Any]]
    inputs: Optional[Dict[str, Any]]
    outputs: Optional[Dict[str, Any]]
    children: Optional[List["V1EventSpan"]]
    kind: Optional[V1EventSpanKind]

    def add_metadata(self, key: str, value: Any) -> None:
        if self.metadata is None:
            self.metadata = {}
        self.metadata[key] = value

    def add_named_result(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        self.inputs = inputs
        self.outputs = outputs

    def add_child_span(self, span: "V1EventSpan") -> None:
        if self.children is None:
            self.children = []
        self.children.append(span)


class V1EventImage(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.IMAGE

    height: Optional[int]
    width: Optional[int]
    colorspace: Optional[int]
    path: Optional[StrictStr]


class V1EventVideo(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.VIDEO

    height: Optional[int]
    width: Optional[int]
    colorspace: Optional[int]
    path: Optional[StrictStr]
    content_type: Optional[StrictStr]


class V1EventDataframe(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.DATAFRAME

    path: Optional[StrictStr]
    content_type: Optional[StrictStr]


class V1EventHistogram(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.HISTOGRAM

    values: Optional[List[float]]
    counts: Optional[List[float]]


class V1EventAudio(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.AUDIO

    sample_rate: Optional[float]
    num_channels: Optional[int]
    length_frames: Optional[int]
    path: Optional[StrictStr]
    content_type: Optional[StrictStr]


class V1EventChartKind(str, PEnum):
    PLOTLY = "plotly"
    BOKEH = "bokeh"
    VEGA = "vega"


class V1EventChart(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.CHART

    kind: Optional[V1EventChartKind]
    figure: Optional[Dict]

    def to_json(self, humanize_values=False, include_kind=False, include_version=False):
        if self.kind == V1EventChartKind.PLOTLY:
            import plotly.tools

            obj = self.to_dict(
                humanize_values=humanize_values,
                include_kind=include_kind,
                include_version=include_version,
            )
            return json.dumps(obj, cls=plotly.utils.PlotlyJSONEncoder)
        # Resume normal serialization
        return super().to_json(
            humanize_values=humanize_values,
            include_kind=include_kind,
            include_version=include_version,
        )


class V1EventCurveKind(str, PEnum):
    ROC = "roc"
    PR = "pr"
    CUSTOM = "custom"


class V1EventCurve(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.CURVE

    kind: Optional[V1EventCurveKind]
    x: Optional[List[float]]
    y: Optional[List[float]]
    annotation: Optional[StrictStr]


class V1EventConfusionMatrix(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.CONFUSION

    x: Optional[List]
    y: Optional[List]
    z: Optional[List]


class V1EventArtifact(BaseSchemaModel):
    _IDENTIFIER = "artifact"

    kind: Optional[V1ArtifactKind]
    path: Optional[StrictStr]


class V1EventModel(BaseSchemaModel):
    _IDENTIFIER = V1ArtifactKind.MODEL

    framework: Optional[StrictStr]
    path: Optional[StrictStr]
    spec: Optional[Dict]


class V1Event(BaseSchemaModel):
    _SEPARATOR = "|"
    _IDENTIFIER = "event"

    timestamp: Optional[datetime.datetime]
    step: Optional[int]
    metric: Optional[float]
    image: Optional[V1EventImage]
    histogram: Optional[V1EventHistogram]
    audio: Optional[V1EventAudio]
    video: Optional[V1EventVideo]
    html: Optional[str]
    text: Optional[str]
    chart: Optional[V1EventChart]
    curve: Optional[V1EventCurve]
    confusion: Optional[V1EventConfusionMatrix]
    artifact: Optional[V1EventArtifact]
    model: Optional[V1EventModel]
    dataframe: Optional[V1EventDataframe]
    span: Optional[V1EventSpan]

    @root_validator(pre=True)
    @skip_partial
    def pre_validate(cls, values):
        v = values.get(V1ArtifactKind.IMAGE)
        get_dict = ConfigParser.parse(Dict)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.IMAGE] = get_dict(
                key=V1ArtifactKind.IMAGE,
                value=v,
            )
        v = values.get(V1ArtifactKind.HISTOGRAM)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.HISTOGRAM] = get_dict(
                key=V1ArtifactKind.HISTOGRAM,
                value=v,
            )
        v = values.get(V1ArtifactKind.AUDIO)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.AUDIO] = get_dict(
                key=V1ArtifactKind.AUDIO,
                value=v,
            )
        v = values.get(V1ArtifactKind.VIDEO)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.VIDEO] = get_dict(
                key=V1ArtifactKind.VIDEO,
                value=v,
            )
        v = values.get(V1ArtifactKind.CHART)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.CHART] = get_dict(
                key=V1ArtifactKind.CHART,
                value=v,
            )
        v = values.get(V1ArtifactKind.CURVE)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.CURVE] = get_dict(
                key=V1ArtifactKind.CURVE,
                value=v,
            )
        v = values.get(V1ArtifactKind.CONFUSION)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.CONFUSION] = get_dict(
                key=V1ArtifactKind.CONFUSION,
                value=v,
            )
        v = values.get(V1ArtifactKind.ARTIFACT)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.ARTIFACT] = get_dict(
                key=V1ArtifactKind.ARTIFACT,
                value=v,
            )
        v = values.get(V1ArtifactKind.MODEL)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.MODEL] = get_dict(
                key=V1ArtifactKind.MODEL,
                value=v,
            )
        v = values.get(V1ArtifactKind.DATAFRAME)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.DATAFRAME] = get_dict(
                key=V1ArtifactKind.DATAFRAME,
                value=v,
            )
        v = values.get(V1ArtifactKind.SPAN)
        if v is not None and not isinstance(v, BaseSchemaModel):
            values[V1ArtifactKind.SPAN] = get_dict(
                key=V1ArtifactKind.SPAN,
                value=v,
            )

        return values

    @root_validator
    @skip_partial
    def validate_event(cls, values):
        count = 0

        def increment(c):
            c += 1
            if c > 1:
                raise ValueError(
                    "An event should have one and only one primitive, found {}.".format(
                        c
                    )
                )
            return c

        if values.get(V1ArtifactKind.METRIC) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.IMAGE) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.HISTOGRAM) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.AUDIO) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.VIDEO) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.HTML) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.TEXT) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.CHART) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.CURVE) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.CONFUSION) is not None:
            count = increment(count)
        if values.get("artifact") is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.MODEL) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.DATAFRAME) is not None:
            count = increment(count)
        if values.get(V1ArtifactKind.SPAN) is not None:
            count = increment(count)

        if count != 1:
            raise ValueError(
                "An event should have one and only one primitive, found {}.".format(
                    count
                )
            )

        return values

    @classmethod
    def make(
        cls,
        step: Optional[int] = None,
        timestamp=None,
        metric: float = None,
        image: V1EventImage = None,
        histogram: V1EventHistogram = None,
        audio: V1EventAudio = None,
        video: V1EventVideo = None,
        html: Optional[str] = None,
        text: Optional[str] = None,
        chart: V1EventChart = None,
        curve: V1EventCurve = None,
        confusion: V1EventConfusionMatrix = None,
        artifact: V1EventArtifact = None,
        model: V1EventModel = None,
        dataframe: V1EventDataframe = None,
        span: V1EventSpan = None,
    ) -> "V1Event":
        if isinstance(timestamp, str):
            try:
                timestamp = parse_datetime(timestamp)
            except Exception as e:
                raise ValueError("Received an invalid timestamp") from e

        return cls(
            timestamp=timestamp if timestamp else now(tzinfo=True),
            step=step,
            metric=metric,
            image=image,
            histogram=histogram,
            audio=audio,
            video=video,
            html=html,
            text=text,
            chart=chart,
            curve=curve,
            confusion=confusion,
            artifact=artifact,
            model=model,
            dataframe=dataframe,
            span=span,
        )

    def get_value(self, dump=True):
        if self.metric is not None:
            return str(self.metric) if dump else self.metric
        if self.image is not None:
            return self.image.to_json() if dump else self.image
        if self.histogram is not None:
            return self.histogram.to_json() if dump else self.histogram
        if self.audio is not None:
            return self.audio.to_json() if dump else self.audio
        if self.video is not None:
            return self.video.to_json() if dump else self.video
        if self.html is not None:
            return self.html
        if self.text is not None:
            return self.text
        if self.chart is not None:
            return self.chart.to_json() if dump else self.chart
        if self.curve is not None:
            return self.curve.to_json() if dump else self.curve
        if self.confusion is not None:
            return self.confusion.to_json() if dump else self.confusion
        if self.artifact is not None:
            return self.artifact.to_json() if dump else self.artifact
        if self.model is not None:
            return self.model.to_json() if dump else self.model
        if self.dataframe is not None:
            return self.dataframe.to_json() if dump else self.dataframe
        if self.span is not None:
            return self.span.to_json() if dump else self.span

    def to_json(
        self,
        humanize_values: bool = False,
        include_kind: bool = False,
        include_version: bool = False,
        kind: Optional[str] = None,
    ) -> str:
        if not kind:
            kind = self._IDENTIFIER
        values = {
            "timestamp": str(self.timestamp),
            "step": self.step,
            kind: self.get_value(dump=True),
        }

        return json.dumps(values)

    def to_csv(self) -> str:
        values = [
            str(self.step) if self.step is not None else "",
            str(self.timestamp) if self.timestamp is not None else "",
            self.get_value(dump=True),
        ]

        return self._SEPARATOR.join(values)


class V1Events:
    ORIENT_CSV = "csv"
    ORIENT_DICT = "dict"

    def __init__(self, kind, name, df):
        self.kind = kind
        self.name = name
        self.df = df

    @staticmethod
    def _read_csv(
        data: str,
        parse_dates: Optional[bool] = True,
        engine: Optional[str] = None,
    ) -> "pandas.DataFrame":
        import pandas as pd

        data = validate_file_or_buffer(data)
        if parse_dates:
            return pd.read_csv(
                data,
                sep=V1Event._SEPARATOR,
                parse_dates=["timestamp"],
                engine=engine,
            )
        else:
            df = pd.read_csv(
                data,
                sep=V1Event._SEPARATOR,
                engine=engine,
            )
            # Pyarrow automatically converts timestamp fields
            if "timestamp" in df.columns:
                df["timestamp"] = df["timestamp"].astype(str)
            return df

    @staticmethod
    def _read_jsonl(
        data: str,
        parse_dates: Optional[bool] = True,
    ) -> "pandas.DataFrame":
        import pandas as pd

        data = validate_file_or_buffer(data)
        if parse_dates:
            df = pd.read_json(
                data,
                lines=True,
            )
        else:
            df = pd.read_json(
                data,
                lines=True,
            )
            # Pyarrow automatically converts timestamp fields
            if "timestamp" in df.columns:
                df["timestamp"] = df["timestamp"].astype(str)
        if not hasattr(df, "step"):
            df["step"] = None
        return df

    @classmethod
    def read(
        cls,
        kind: str,
        name: str,
        data: Union[str, Dict],
        parse_dates: Optional[bool] = True,
        engine: Optional[str] = None,
    ) -> "V1Events":
        import pandas as pd

        if isinstance(data, str):
            error = None
            if V1ArtifactKind.is_jsonl_file_event(kind):
                try:
                    df = cls._read_jsonl(data=data, parse_dates=parse_dates)
                except Exception as e:
                    error = e
            if not V1ArtifactKind.is_jsonl_file_event(kind) or error:
                df = cls._read_csv(data=data, parse_dates=parse_dates, engine=engine)
        elif isinstance(data, dict):
            df = pd.DataFrame.from_dict(data)
        else:
            raise ValueError(
                "V1Events received an unsupported value type: {}".format(type(data))
            )

        return cls(name=name, kind=kind, df=df)

    def to_dict(self, orient: str = "list") -> Dict:
        import numpy as np

        return self.df.replace({np.nan: None}).to_dict(orient=orient)

    def get_event_at(self, index):
        event = self.df.iloc[index].to_dict()
        event["timestamp"] = event["timestamp"].isoformat()
        event["step"] = sanitize_np_types(event["step"])
        return V1Event.from_dict(event)

    def _get_step_summary(self) -> Optional[Dict]:
        _count = self.df.step.count()
        if _count == 0:
            return None

        return {
            "count": sanitize_np_types(_count),
            "min": sanitize_np_types(self.df.step.iloc[0]),
            "max": sanitize_np_types(self.df.step.iloc[-1]),
        }

    def _get_ts_summary(self) -> Optional[Dict]:
        _count = self.df.timestamp.count()
        if _count == 0:
            return None

        try:
            return {
                "min": self.df.timestamp.iloc[0].isoformat(),
                "max": self.df.timestamp.iloc[-1].isoformat(),
            }
        except Exception as e:
            logger.debug("Received an error while parsing timestamps: %s", e)
            return None

    def get_summary(self) -> Dict:
        summary = {"is_event": True}
        step_summary = self._get_step_summary()
        if step_summary:
            summary["step"] = step_summary

        ts_summary = self._get_ts_summary()
        if ts_summary:
            summary["timestamp"] = ts_summary

        if self.kind == V1ArtifactKind.METRIC:
            summary[self.kind] = {
                k: sanitize_np_types(v)
                for k, v in self.df.metric.describe().to_dict().items()
            }
            summary[self.kind]["last"] = sanitize_np_types(self.df.metric.iloc[-1])

        return summary


class LoggedEventSpec(namedtuple("LoggedEventSpec", "name kind event")):
    pass


class LoggedEventListSpec(namedtuple("LoggedEventListSpec", "name kind events")):
    def get_csv_header(self) -> str:
        return V1Event._SEPARATOR.join(["step", "timestamp", self.kind])

    def get_csv_events(self) -> str:
        events = ["\n{}".format(e.to_csv()) for e in self.events]
        return "".join(events)

    def get_jsonl_events(self) -> str:
        events = ["\n{}".format(e.to_json(kind=self.kind)) for e in self.events]
        return "".join(events)

    def empty_events(self):
        self.events[:] = []

    def to_dict(self):
        return {
            "name": self.name,
            "kind": self.kind,
            "events": [e.to_dict() for e in self.events],
        }

    @classmethod
    def from_dict(cls, value: Mapping) -> "LoggedEventListSpec":
        return cls(
            name=value.get("name"),
            kind=value.get("kind"),
            events=[V1Event.from_dict(e) for e in value.get("events", [])],
        )



import logging
import os
import socket

from clipped.utils.dates import to_datetime
from clipped.utils.env import get_user

from polyaxon import settings
from polyaxon._env_vars.keys import ENV_KEYS_K8S_NODE_NAME, ENV_KEYS_K8S_POD_ID
from traceml.logging.schemas import V1Log


class LogStreamHandler(logging.Handler):
    def __init__(self, add_logs, **kwargs):
        self._add_logs = add_logs
        self._container = socket.gethostname()
        self._node = os.environ.get(ENV_KEYS_K8S_NODE_NAME, "local")
        self._pod = os.environ.get(ENV_KEYS_K8S_POD_ID, get_user())
        log_level = settings.CLIENT_CONFIG.log_level
        if log_level and isinstance(log_level, str):
            log_level = log_level.upper()
        super().__init__(
            level=kwargs.get("level", log_level or logging.NOTSET),
        )

    def set_add_logs(self, add_logs):
        self._add_logs = add_logs

    def can_record(self, record):
        return not (
            record.name == "polyaxon"
            or record.name == "traceml"
            or record.name == "polyaxon.cli"
            or record.name.startswith("polyaxon")
            or record.name.startswith("traceml")
        )

    def format_record(self, record):
        message = ""
        if record.msg:
            message = record.msg
        if record.args:
            message %= record.args
        return V1Log.process_log_line(
            value=message,
            timestamp=to_datetime(record.created),
            node=self._node,
            pod=self._pod,
            container=self._container,
        )

    def emit(self, record):  # pylint:disable=inconsistent-return-statements
        if not self.can_record(record):
            return
        try:
            return self._add_logs(self.format_record(record))
        except Exception:  # noqa
            pass


class LogStreamWriter:
    def __init__(self, logger, log_level, channel):
        self._logger = logger
        self._log_level = log_level
        self._channel = channel

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            if line != "\n":
                self._logger.log(self._log_level, line.rstrip())

    def flush(self):
        self._channel.flush()



from typing import Optional

from clipped.utils.np import to_np
from clipped.utils.paths import check_or_create_path, copy_file_path

from polyaxon._constants.globals import UNKNOWN
from traceml.events import V1EventAudio
from traceml.logger import logger
from traceml.processors.errors import NUMPY_ERROR_MESSAGE

try:
    import numpy as np
except ImportError:
    np = None


def audio_path(
    from_path: str,
    asset_path: str,
    content_type=None,
    asset_rel_path: Optional[str] = None,
) -> V1EventAudio:
    copy_file_path(from_path, asset_path)
    return V1EventAudio(path=asset_rel_path or asset_path, content_type=content_type)


def audio(
    asset_path: str, tensor, sample_rate=44100, asset_rel_path: Optional[str] = None
):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    tensor = to_np(tensor)
    tensor = tensor.squeeze()
    if abs(tensor).max() > 1:
        print("warning: audio amplitude out of range, auto clipped.")
        tensor = tensor.clip(-1, 1)
    assert tensor.ndim == 1, "input tensor should be 1 dimensional."

    tensor_list = [int(32767.0 * x) for x in tensor]

    import struct
    import wave

    check_or_create_path(asset_path, is_dir=False)

    wave_write = wave.open(asset_path, "wb")
    wave_write.setnchannels(1)
    wave_write.setsampwidth(2)
    wave_write.setframerate(sample_rate)
    tensor_enc = b""
    for v in tensor_list:
        tensor_enc += struct.pack("<h", v)

    wave_write.writeframes(tensor_enc)
    wave_write.close()
    return V1EventAudio(
        sample_rate=sample_rate,
        num_channels=1,
        length_frames=len(tensor_list),
        path=asset_rel_path or asset_path,
        content_type="audio/wav",
    )



import uuid

from typing import Dict, List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.types.uuids import UUIDStr

from polyaxon._schemas.base import BaseSchemaModel
from traceml.artifacts.enums import V1ArtifactKind


class V1RunArtifact(BaseSchemaModel):
    _IDENTIFIER = "artifact"

    name: Optional[StrictStr]
    kind: Optional[V1ArtifactKind]
    path: Optional[StrictStr]
    state: Optional[UUIDStr]
    summary: Optional[Dict]
    meta_info: Optional[Dict]
    run: Optional[UUIDStr]
    connection: Optional[StrictStr]
    is_input: Optional[bool]

    @classmethod
    def from_model(cls, model):
        return cls(
            name=model.name,
            kind=model.kind,
            path=model.path,
            state=model.state,
            summary=model.summary,
            # connection=model.connection,  # TODO: enable
        )

    def get_state(self, namespace: uuid.UUID):
        if self.state:
            return self.state
        summary = self.summary or {}
        content = str(summary)
        if not summary.get("hash") and self.path:
            content += self.path
        return uuid.uuid5(namespace, content)


class V1RunArtifacts(BaseSchemaModel):
    _IDENTIFIER = "artifacts"

    artifacts: Optional[List[V1RunArtifact]]



from traceml.processors.events_processors.events_artifacts_processors import (
    artifact_path,
)
from traceml.processors.events_processors.events_audio_processors import (
    audio,
    audio_path,
)
from traceml.processors.events_processors.events_charts_processors import (
    altair_chart,
    bokeh_chart,
    mpl_plotly_chart,
    plotly_chart,
)
from traceml.processors.events_processors.events_image_processors import (
    convert_to_HWC,
    draw_boxes,
    encoded_image,
    ensure_matplotlib_figure,
    figure_to_image,
    figures_to_images,
    image,
    image_boxes,
    image_path,
    make_grid,
    make_image,
    save_image,
)
from traceml.processors.events_processors.events_metrics_processors import (
    confusion_matrix,
    curve,
    histogram,
    metric,
    metrics_dict_to_list,
    np_histogram,
    pr_curve,
    roc_auc_curve,
    sklearn_pr_curve,
    sklearn_roc_auc_curve,
)
from traceml.processors.events_processors.events_models_processors import (
    model_path,
    model_to_str,
)
from traceml.processors.events_processors.events_tables_processors import dataframe_path
from traceml.processors.events_processors.events_video_processors import (
    make_video,
    prepare_video,
    video,
    video_path,
)



import math
import numpy as np
import pandas as pd

from typing import Dict, List, Optional, Tuple, Union

from clipped.utils.np import sanitize_np_types


def clean_duplicates(
    params: pd.DataFrame, metrics: pd.DataFrame
) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    duplicate_ids = metrics.duplicated()
    params_df = params[~duplicate_ids]
    metrics_df = metrics[~duplicate_ids]
    if params.empty or metrics.empty:
        return None

    params_df = pd.get_dummies(params_df)
    params_df = params_df.loc[:, ~params_df.columns.duplicated()]
    return params_df, metrics_df


def clean_values(
    params: List[Dict], metrics: List[Union[int, float]]
) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    if not metrics or not params:
        return None

    for m in metrics:
        if not isinstance(m, (int, float)):
            return None

    metrics_df = pd.DataFrame(metrics)
    if metrics_df.isnull().values.any():
        return None

    params_df = pd.DataFrame.from_records(params).replace(r"^\s*$", np.nan, regex=True)
    for col in params_df.columns:
        if not params_df[col].isnull().sum() == len(params_df[col]):
            if params_df[col].dtype == "object":
                params_df[col].fillna("NAN", inplace=True)
                params_df[col].fillna("NAN", inplace=True)
                params_df[col] = params_df[col].astype("category")
            elif params_df[col].dtype == "float64" or params_df[col].dtype == "int64":
                params_df[col].fillna(params_df[col].mean(), inplace=True)
            else:
                print("Unexpected Column type: {}".format(params_df[col].dtype))
        else:
            if params_df[col].dtype == "object":
                params_df[col] = "NAN"
            elif params_df[col].dtype == "float64" or params_df[col].dtype == "int64":
                params_df[col] = 0

    return clean_duplicates(params_df, metrics_df)


def _get_value(x):
    if x is None or math.isnan(x):
        return None
    return round(sanitize_np_types(x), 3)


def calculate_importance_correlation(
    params: List[Dict], metrics: List[Union[int, float]]
):
    values = clean_values(params, metrics)
    if not values:
        return None
    params_df, metrics_df = values

    corr_list = params_df.corrwith(metrics_df[0])

    from sklearn.ensemble import ExtraTreesRegressor

    forest = ExtraTreesRegressor(n_estimators=250, random_state=0)
    forest.fit(params_df, metrics_df[0])
    feature_importances = forest.feature_importances_

    results = {}
    for i, name in enumerate(params_df.columns):
        results[name] = {
            "importance": _get_value(feature_importances[i]),
            "correlation": _get_value(corr_list[name]),
        }
    return results



NUMPY_ERROR_MESSAGE = "numpy is required for this tracking operation."
PANDAS_ERROR_MESSAGE = "pandas is required for this tracking operation."
PIL_ERROR_MESSAGE = "PIL/Pillow is required for this tracking operation."
MOVIEPY_ERROR_MESSAGE = "moviepy is required for this tracking operation."
MATPLOTLIB_ERROR_MESSAGE = "matplotlib is required for this tracking operation."
PLOTLY_ERROR_MESSAGE = "plotly is required for this tracking operation."
BOKEH_ERROR_MESSAGE = "bokeh is required for this tracking operation."
SKLEARN_ERROR_MESSAGE = "sklearn is required for this tracking operation."



import numpy as np
import pandas as pd

from traceml.processors import df_processors


class DataFrameSummary:
    ALL = "ALL"
    INCLUDE = "INCLUDE"
    EXCLUDE = "EXCLUDE"

    TYPE_BOOL = "bool"
    TYPE_NUMERIC = "numeric"
    TYPE_DATE = "date"
    TYPE_CATEGORICAL = "categorical"
    TYPE_CONSTANT = "constant"
    TYPE_UNIQUE = "unique"

    def __init__(self, df, plot=False):
        self.df = df
        self.length = len(df)
        self._columns_stats = None
        self.plot = plot

    def __getitem__(self, column):
        if isinstance(column, str) and df_processors.df_has_column(
            df=self.df, column=column
        ):
            return df_processors.get_df_column_summary(
                df=self.df,
                column=column,
                columns_stats=self.columns_stats,
                df_length=self.length,
                plot=self.plot,
            )

        if isinstance(column, int) and column < self.df.shape[1]:
            return df_processors.get_df_column_summary(
                df=self.df,
                column=self.df.columns[column],
                columns_stats=self.columns_stats,
                df_length=self.length,
                plot=self.plot,
            )

        if isinstance(column, (tuple, list)):
            error_keys = [
                k
                for k in column
                if not df_processors.df_has_column(df=self.df, column=k)
            ]
            if len(error_keys) > 0:
                raise KeyError(", ".join(error_keys))
            return self.df[list(column)].values

        if isinstance(column, pd.Index):
            error_keys = [
                k
                for k in column.values
                if not df_processors.df_has_column(df=self.df, column=k)
            ]
            if len(error_keys) > 0:
                raise KeyError(", ".join(error_keys))
            return self.df[column].values

        if isinstance(column, np.ndarray):
            error_keys = [
                k
                for k in column
                if not df_processors.df_has_column(df=self.df, column=k)
            ]
            if len(error_keys) > 0:
                raise KeyError(", ".join(error_keys))
            return self.df[column].values

        raise KeyError(column)

    @property
    def columns_stats(self):
        if self._columns_stats:
            return self._columns_stats
        self._columns_stats = df_processors.get_df_column_stats(self.df)
        return self._columns_stats

    @property
    def columns_types(self):
        return df_processors.get_df_columns_types(self.columns_stats)

    def summary(self):
        return pd.concat([self.df.describe(), self.columns_stats], sort=True)[
            self.df.columns
        ]

    """ Column summaries """

    @property
    def constants(self):
        return self.df.columns[self.columns_stats.loc["types"] == "constant"]

    @property
    def categoricals(self):
        return self.df.columns[self.columns_stats.loc["types"] == "categorical"]

    @property
    def numerics(self):
        return self.df.columns[self.columns_stats.loc["types"] == "numeric"]

    @property
    def uniques(self):
        return self.df.columns[self.columns_stats.loc["types"] == "unique"]

    @property
    def bools(self):
        return self.df.columns[self.columns_stats.loc["types"] == "bool"]

    @property
    def missing_frac(self):
        return self.columns_stats.loc["missing"].apply(lambda x: float(x) / self.length)

    def get_columns(self, df, usage, columns=None):
        """
        Returns a `data_frame.columns`.
        :param df: dataframe to select columns from
        :param usage: should be a value from [ALL, INCLUDE, EXCLUDE].
                            this value only makes sense if attr `columns` is also set.
                            otherwise, should be used with default value ALL.
        :param columns: * if `usage` is all, this value is not used.
                        * if `usage` is INCLUDE, the `df` is restricted to the intersection
                          between `columns` and the `df.columns`
                        * if usage is EXCLUDE, returns the `df.columns` excluding these `columns`
        :return: `data_frame` columns, excluding `target_column` and `id_column` if given.
                 `data_frame` columns, including/excluding the `columns` depending on `usage`.
        """
        columns_excluded = pd.Index([])
        columns_included = df.columns

        if usage == self.INCLUDE:
            try:
                columns_included = columns_included.intersection(pd.Index(columns))
            except TypeError:
                pass
        elif usage == self.EXCLUDE:
            try:
                columns_excluded = columns_excluded.union(pd.Index(columns))
            except TypeError:
                pass

        columns_included = columns_included.difference(columns_excluded)
        return columns_included.intersection(df.columns)
